import asyncio
import json
import time
import functools
from ring_doorbell import Ring, Auth
from oauthlib.oauth2 import MissingTokenError, AccessDeniedError

try:
    # Python 3.7
    from asyncio.futures import CancelledError
except ImportError:
    # Python 3.8+
    from asyncio import CancelledError

from .. import fhem
from .. import utils
from fhempy.lib.generic import FhemModule


class ring(FhemModule):
    def __init__(self, logger):
        super().__init__(logger)
        self.loop = asyncio.get_event_loop()
        self._username = None
        self._password = ""
        self._token = ""
        self._2facode = None
        self._attr_dingPollInterval = 5
        self._attr_deviceUpdateInterval = 300
        self._history = []
        self._rdevice = None
        self._lastrecording_url = ""
        self._livestreamjson = ""
        self._login_lock = asyncio.Lock()
        self._snapshot = None
        self._attr_list = {
            "deviceUpdateInterval": {"default": 300, "format": "int"},
            "dingPollInterval": {"default": 2, "format": "int"},
        }
        self.set_attr_config(self._attr_list)
        self.set_list_conf = {
            "password": {"args": ["password"]},
            "2fa_code": {"args": ["2facode"]},
        }
        self.set_set_config(self.set_list_conf)
        return

    async def token_updated(self, token):
        self._token = token
        encrypted_token = utils.encrypt_string(
            json.dumps(token), self._reading_encryption_key
        )
        await fhem.readingsSingleUpdate(self.hash, "token", encrypted_token, 1)

    # FHEM FUNCTION
    async def Define(self, hash, args, argsh):
        await super().Define(hash, args, argsh)
        if len(args) < 5:
            return (
                "Usage: define rrring PythonModule ring <USERNAME> <RING_DEVICE_NAME>"
            )
        self._username = args[3]
        self._rdevname = args[4]
        self._reading_encryption_key = await fhem.getUniqueId(hash)
        # ring service
        self._ring = None
        # ring doorbell/chime/... device
        self._rdevice = None
        self.hash["USERNAME"] = args[3]
        self.hash["RINGDEVICE"] = args[4]
        self.create_async_task(self.ring_login())

    async def ring_login(self):
        async with self._login_lock:
            if self._token == "":
                token_reading = await fhem.ReadingsVal(self.hash["NAME"], "token", "")
                if token_reading != "":
                    token_reading = utils.decrypt_string(
                        token_reading, self._reading_encryption_key
                    )
                    self._token = json.loads(token_reading)
            if self._password == "":
                self._password = await fhem.ReadingsVal(
                    self.hash["NAME"], "password", ""
                )
                if self._password != "":
                    self._password = utils.decrypt_string(
                        self._password, self._reading_encryption_key
                    )
            try:
                ret = await utils.run_blocking(functools.partial(self.blocking_login))
                if ret:
                    await fhem.readingsSingleUpdate(self.hash, "state", ret, 1)
                    return
                await fhem.readingsSingleUpdate(self.hash, "state", "connected", 1)
                # start update loop, we are already in a task, therefore no need to create
                await self.update_loop()
            except:
                await fhem.readingsSingleUpdate(self.hash, "state", "Login failed", 1)
                self.logger.error("Login failed")

    async def update_loop(self):
        try:
            await utils.run_blocking(functools.partial(self._ring.update_data))
            devices = await utils.run_blocking(functools.partial(self._ring.devices))
            for dev_type in devices:
                for dev in devices[dev_type]:
                    if dev.name == self._rdevname:
                        self._rdevice = dev

            if self._rdevice is None:
                await fhem.readingsSingleUpdate(
                    self.hash, "state", "device not found", 1
                )
                return

            if self._rdevice is not None and self._rdevice.has_capability("volume"):
                self.set_list_conf["volume"] = {
                    "args": ["volume"],
                    "params": {"volume": {"format": "int"}},
                    "options": "slider,0,1,10",
                }
            self.set_set_config(self.set_list_conf)

            await self.update_readings()

            self.create_async_task(self.update_dings_loop())

            while True:
                try:
                    await utils.run_blocking(functools.partial(self.poll_device))
                    await self.update_readings()
                    # handle history
                    if len(self._history) > 0:
                        i = 1
                        for event in self._history:
                            await self.update_history_readings(event, i)
                            i += 1
                except:
                    self.logger.exception("Failed to poll devices")
                await asyncio.sleep(self._attr_deviceUpdateInterval)
        except CancelledError:
            pass
        except:
            self.logger.exception("Failed to update devices")

    async def update_dings_loop(self):
        alert_active = 0
        while True:
            try:
                await utils.run_blocking(functools.partial(self.poll_dings))
                # handle alerts
                alerts = self._ring.active_alerts()
                self.logger.debug("Received dings: " + str(alerts))
                if len(alerts) > 0:
                    for alert in alerts:
                        if alert["doorbot_id"] == self._rdevice.id:
                            alert_active = 1
                            await self.update_alert_readings(alert)
                elif alert_active == 1:
                    alert_active = 0
                    await fhem.readingsSingleUpdateIfChanged(
                        self.hash, "state", "connected", 1
                    )
                    await utils.run_blocking(functools.partial(self.poll_device))
                    await self.update_readings()
            except:
                self.logger.exception("Failed to poll dings...")
                await utils.run_blocking(functools.partial(self.blocking_login))
            await asyncio.sleep(self._attr_dingPollInterval)

    async def update_alert_readings(self, alert):
        await fhem.readingsBeginUpdate(self.hash)
        await fhem.readingsBulkUpdate(self.hash, "alert_id", alert["id"])
        await fhem.readingsBulkUpdate(self.hash, "alert_kind", alert["kind"])
        await fhem.readingsBulkUpdate(self.hash, "alert_sip_to", alert["sip_to"])
        await fhem.readingsBulkUpdate(self.hash, "alert_sip_token", alert["sip_token"])
        await fhem.readingsBulkUpdate(
            self.hash, "alert_doorbot_id", alert["doorbot_id"]
        )
        await fhem.readingsBulkUpdate(self.hash, "state", alert["kind"])
        await fhem.readingsEndUpdate(self.hash, 1)

    async def update_history_readings(self, event, idx):
        await fhem.readingsBeginUpdate(self.hash)
        await fhem.readingsBulkUpdateIfChanged(
            self.hash, "history_" + str(idx) + "_id", event["id"]
        )
        await fhem.readingsBulkUpdateIfChanged(
            self.hash, "history_" + str(idx) + "_kind", event["kind"]
        )
        await fhem.readingsBulkUpdateIfChanged(
            self.hash, "history_" + str(idx) + "_answered", event["answered"]
        )
        await fhem.readingsBulkUpdateIfChanged(
            self.hash, "history_" + str(idx) + "_created_at", event["created_at"]
        )
        await fhem.readingsEndUpdate(self.hash, 1)

    async def update_readings(self):
        await fhem.readingsBeginUpdate(self.hash)
        try:
            await fhem.readingsBulkUpdateIfChanged(
                self.hash, "address", self._rdevice.address
            )
            await fhem.readingsBulkUpdateIfChanged(
                self.hash, "family", self._rdevice.family
            )
            await fhem.readingsBulkUpdateIfChanged(
                self.hash, "device_id", self._rdevice.device_id
            )
            await fhem.readingsBulkUpdateIfChanged(self.hash, "id", self._rdevice.id)
            await fhem.readingsBulkUpdateIfChanged(
                self.hash, "model", self._rdevice.model
            )
            await fhem.readingsBulkUpdateIfChanged(
                self.hash, "firmware", self._rdevice.firmware
            )
            await fhem.readingsBulkUpdateIfChanged(
                self.hash, "latitude", self._rdevice.latitude
            )
            await fhem.readingsBulkUpdateIfChanged(
                self.hash, "longitude", self._rdevice.longitude
            )
            await fhem.readingsBulkUpdateIfChanged(
                self.hash, "kind", self._rdevice.kind
            )
            await fhem.readingsBulkUpdateIfChanged(
                self.hash, "name", self._rdevice.name
            )
            await fhem.readingsBulkUpdateIfChanged(
                self.hash, "timezone", self._rdevice.timezone
            )
            await fhem.readingsBulkUpdateIfChanged(
                self.hash, "wifi_name", self._rdevice.wifi_name
            )
            await fhem.readingsBulkUpdateIfChanged(
                self.hash, "wifi_signal_strength", self._rdevice.wifi_signal_strength
            )
            await fhem.readingsBulkUpdateIfChanged(
                self.hash, "wifi_signal_category", self._rdevice.wifi_signal_category
            )
            await self.update_if_available("battery_life")
            await self.update_if_available("existing_doorbell_type")
            await self.update_if_available("existing_doorbell_type_enabled")
            await self.update_if_available("existing_doorbell_type_duration")
            await self.update_if_available("subscribed")
            await self.update_if_available("subscribed_motion")
            await self.update_if_available("has_subscription")
            await self.update_if_available("volume")
            await self.update_if_available("connection_status")
            if (
                self._rdevice.family == "doorbots"
                or self._rdevice.family == "authorized_doorbots"
            ):
                await self.update_if_available("last_recording_id")
                await fhem.readingsBulkUpdateIfChanged(
                    self.hash, "last_recording_url", self._lastrecording_url
                )
                if self._livestreamjson != "":
                    await fhem.readingsBulkUpdateIfChanged(
                        self.hash, "livestream_json", json.dumps(self._livestreamjson)
                    )
                # if self._snapshot:
                #     snapshot = '<html><img src="data:image/png,' + self._snapshot + '"/></html>'
                #     await fhem.readingsBulkUpdateIfChanged(self.hash, "snapshot", snapshot)
        except:
            self.logger.exception(
                "Failed to update readings, please report here: https://forum.fhem.de/index.php/topic,117381"
            )
        await fhem.readingsEndUpdate(self.hash, 1)

    async def update_if_available(self, reading):
        if hasattr(self._rdevice, reading):
            await fhem.readingsBulkUpdateIfChanged(
                self.hash, reading, getattr(self._rdevice, reading)
            )

    def poll_dings(self):
        self._ring.update_dings()

    def poll_device(self):
        self._rdevice.update_health_data()
        self._history = []
        if (
            self._rdevice.family == "doorbots"
            or self._rdevice.family == "authorized_doorbots"
        ):
            # disable it, as it creates a lot of history entries
            # self._livestreamjson = self._rdevice.live_streaming_json
            for event in self._rdevice.history(limit=5):
                self._history.append(event)
            self._lastrecording_url = self._rdevice.recording_url(
                self._rdevice.last_recording_id
            )
            if self._lastrecording_url is False:
                self.blocking_login()
            # self._snapshot = self._rdevice.get_snapshot()

    def blocking_login(self):
        def token_updater(token):
            asyncio.run_coroutine_threadsafe(
                self.token_updated(token), self.loop
            ).result()

        if self._token != "":
            self._auth = Auth("MyProject/1.0", self._token, token_updater)
        else:
            if self._password != "":
                self._auth = Auth("MyProject/1.0", None, token_updater)
                if self._2facode:
                    self._auth.fetch_token(
                        self._username, self._password, self._2facode
                    )
                else:
                    try:
                        self._auth.fetch_token(self._username, self._password)
                    except AccessDeniedError:
                        return "AccessDenied: wrong username or password"
                    except MissingTokenError:
                        return "please set 2fa_code"
            else:
                return "please set password"
        self._ring = Ring(self._auth)

    async def set_volume(self, hash, params):
        new_vol = params["volume"]
        self.create_async_task(self.set_volume_task(new_vol))

    async def set_volume_task(self, new_volume):
        await utils.run_blocking(
            functools.partial(self.set_volume_blocking, new_volume)
        )
        await self.update_readings()

    def set_volume_blocking(self, new_volume):
        self._rdevice.volume = new_volume

    async def set_password(self, hash, params):
        self._password = params["password"]
        encrypted_password = utils.encrypt_string(
            self._password, self._reading_encryption_key
        )
        await fhem.readingsSingleUpdateIfChanged(
            self.hash, "password", encrypted_password, 1
        )
        self.create_async_task(self.ring_login())

    async def set_2fa_code(self, hash, params):
        self._2facode = params["2facode"]
        self.create_async_task(self.ring_login())



from fhempy.lib.xiaomi_gateway3_device.devices.base import BaseDevice
from fhempy.lib import fhem, utils
import functools


class Gateway(BaseDevice):
    def __init__(self, logger, gateway):
        super().__init__(logger, gateway)
        set_config = {
            "start_pairing": {},
            "stop_pairing": {},
            "firmware_update": {"args": ["block"], "options": "block,allow"},
        }
        self.set_set_config(set_config)

    async def set_firmware_update(self, hash, params):
        self.create_async_task(self.lock_firmware(params["block"] == "block"))

    async def lock_firmware(self, enable):
        locked = await utils.run_blocking(
            functools.partial(self._gateway.gateway3.lock_firmware, enable)
        )
        await fhem.readingsSingleUpdateIfChanged(
            self.hash, "firmware_lock", str(locked), 1
        )

    async def set_start_pairing(self, hash, params):
        self._gateway.gateway3.miio.send(
            "miIO.zb_start_provision",
            {
                "dev_type": 0,
                "duration": 60,
                "method": 0,
                "model": "lumi.sensor_switch.v2",
                "pid": 62,
            },
        )

    async def set_stop_pairing(self, hash, params):
        self._gateway.gateway3.miio.send("miIO.zb_end_provision", {"code": -1})



"""
The base logic was taken from project https://github.com/squachen/micloud

I had to rewrite the code to work asynchronously and handle timeouts for
requests to the cloud.

MIT License

Copyright (c) 2020 Sammy Svensson

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
import asyncio
import base64
import hashlib
import hmac
import json
import logging
import os
import random
import string
import time

from aiohttp import ClientSession

_LOGGER = logging.getLogger(__name__)

SERVERS = ['cn', 'de', 'i2', 'ru', 'sg', 'us']
UA = "Android-7.1.1-1.0.0-ONEPLUS A3010-136-%s APP/xiaomi.smarthome APPV/62830"


class MiCloud:
    auth = None

    def __init__(self, session: ClientSession):
        self.session = session
        self.device_id = get_random_string(16)

    async def login(self, username: str, password: str):
        try:
            payload = await self._login_step1()
            data = await self._login_step2(username, password, payload)
            if not data['location']:
                return False

            token = await self._login_step3(data['location'])

            self.auth = {
                'user_id': data['userId'],
                'ssecurity': data['ssecurity'],
                'service_token': token
            }

            return True

        except Exception as e:
            _LOGGER.exception(f"Can't login to Mi Cloud: {e}")
            return False

    async def _login_step1(self):
        r = await self.session.get(
            'https://account.xiaomi.com/pass/serviceLogin',
            cookies={'sdkVersion': '3.8.6', 'deviceId': self.device_id},
            headers={'User-Agent': UA % self.device_id},
            params={'sid': 'xiaomiio', '_json': 'true'})
        raw = await r.read()
        _LOGGER.debug(f"MiCloud step1")
        resp: dict = json.loads(raw[11:])
        return {k: v for k, v in resp.items()
                if k in ('sid', 'qs', 'callback', '_sign')}

    async def _login_step2(self, username: str, password: str, payload: dict):
        payload['user'] = username
        payload['hash'] = hashlib.md5(password.encode()).hexdigest().upper()

        r = await self.session.post(
            'https://account.xiaomi.com/pass/serviceLoginAuth2',
            cookies={'sdkVersion': '3.8.6', 'deviceId': self.device_id},
            data=payload,
            headers={'User-Agent': UA % self.device_id},
            params={'_json': 'true'})
        raw = await r.read()
        _LOGGER.debug(f"MiCloud step2")
        resp = json.loads(raw[11:])
        return resp

    async def _login_step3(self, location):
        r = await self.session.get(location, headers={'User-Agent': UA})
        service_token = r.cookies['serviceToken'].value
        _LOGGER.debug(f"MiCloud step3")
        return service_token

    async def get_total_devices(self, servers: list):
        total = []
        for server in servers:
            devices = await self.get_devices(server)
            if devices is None:
                return None
            total += devices
        return total

    async def get_devices(self, server: str):
        assert server in SERVERS, "Wrong server: " + server
        baseurl = 'https://api.io.mi.com/app' if server == 'cn' \
            else f"https://{server}.api.io.mi.com/app"

        url = '/home/device_list'
        data = '{"getVirtualModel":false,"getHuamiDevices":0}'

        nonce = gen_nonce()
        signed_nonce = gen_signed_nonce(self.auth['ssecurity'], nonce)
        signature = gen_signature(url, signed_nonce, nonce, data)

        try:
            r = await self.session.post(baseurl + url, cookies={
                'userId': self.auth['user_id'],
                'serviceToken': self.auth['service_token'],
                'locale': 'en_US'
            }, headers={
                'User-Agent': UA,
                'x-xiaomi-protocal-flag-cli': 'PROTOCAL-HTTP2'
            }, data={
                'signature': signature,
                '_nonce': nonce,
                'data': data
            }, timeout=10)

            resp = await r.json(content_type=None)
            assert resp['code'] == 0, resp
            return resp['result']['list']

        except asyncio.TimeoutError:
            _LOGGER.error("Timeout while loading MiCloud device list")
        except:
            _LOGGER.exception(f"Can't load devices list")

        return None


def get_random_string(length: int):
    seq = string.ascii_uppercase + string.digits
    return ''.join((random.choice(seq) for _ in range(length)))


def gen_nonce() -> str:
    """Time based nonce."""
    nonce = os.urandom(8) + int(time.time() / 60).to_bytes(4, 'big')
    return base64.b64encode(nonce).decode()


def gen_signed_nonce(ssecret: str, nonce: str) -> str:
    """Nonce signed with ssecret."""
    m = hashlib.sha256()
    m.update(base64.b64decode(ssecret))
    m.update(base64.b64decode(nonce))
    return base64.b64encode(m.digest()).decode()


def gen_signature(url: str, signed_nonce: str, nonce: str, data: str) -> str:
    """Request signature based on url, signed_nonce, nonce and data."""
    sign = '&'.join([url, signed_nonce, nonce, 'data=' + data])
    signature = hmac.new(key=base64.b64decode(signed_nonce),
                         msg=sign.encode(),
                         digestmod=hashlib.sha256).digest()
    return base64.b64encode(signature).decode()



import logging

logger = logging.getLogger(__name__)
logger.setLevel(logging.ERROR)

function_active = []
update_locks = {}
wsconnection = None

readings = {}
attributes = {}
internals = {}
command_define = ""
command_attr = ""


def mock_module(mocker):
    mocker.patch("fhempy.lib.fhem.updateConnection", updateConnection)
    mocker.patch("fhempy.lib.fhem.setFunctionActive", setFunctionActive)
    mocker.patch("fhempy.lib.fhem.getUniqueId", getUniqueId)
    mocker.patch("fhempy.lib.fhem.ReadingsVal", ReadingsVal)
    mocker.patch("fhempy.lib.fhem.AttrVal", AttrVal)
    mocker.patch("fhempy.lib.fhem.InternalVal", InternalVal)
    mocker.patch("fhempy.lib.fhem.addToDevAttrList", addToDevAttrList)
    mocker.patch("fhempy.lib.fhem.setDevAttrList", setDevAttrList)
    mocker.patch("fhempy.lib.fhem.readingsBeginUpdate", readingsBeginUpdate)
    mocker.patch(
        "fhempy.lib.fhem.readingsBulkUpdateIfChanged", readingsBulkUpdateIfChanged
    )
    mocker.patch("fhempy.lib.fhem.readingsBulkUpdate", readingsBulkUpdate)
    mocker.patch("fhempy.lib.fhem.readingsEndUpdate", readingsEndUpdate)
    mocker.patch("fhempy.lib.fhem.readingsSingleUpdate", readingsSingleUpdate)
    mocker.patch(
        "fhempy.lib.fhem.readingsSingleUpdateIfChanged", readingsSingleUpdateIfChanged
    )
    mocker.patch("fhempy.lib.fhem.CommandDefine", CommandDefine)
    mocker.patch("fhempy.lib.fhem.CommandAttr", CommandAttr)
    mocker.patch("fhempy.lib.fhem.CommandDeleteReading", CommandDeleteReading)
    mocker.patch("fhempy.lib.fhem.checkIfDeviceExists", checkIfDeviceExists)
    mocker.patch("fhempy.lib.fhem.convertValue", convertValue)
    mocker.patch("fhempy.lib.fhem.send_version", send_version)


def updateConnection(ws):
    return


def setFunctionActive(hash):
    function_active.append(hash["NAME"])


def setFunctionInactive(hash):
    element = function_active.pop()
    if element != hash["NAME"]:
        logger.error(
            f"Set wrong function inactive, tried {hash['NAME']}, current function_active: {function_active},{element}"
        )


async def getUniqueId(hash):
    return "12345678901234567890123456789012"


async def ReadingsVal(name, reading, default):
    if name in readings and reading in readings[name]:
        return readings[name][reading]
    return default


async def AttrVal(name, attr, default):
    if name in attributes and attr in attributes[name]:
        return attributes[name][attr]
    return default


async def InternalVal(name, internal, default):
    if name in internals and internal in internals[name]:
        return internals[name][internal]
    return default


async def addToDevAttrList(name, attr_list):
    return


async def setDevAttrList(name, attr_list):
    return


async def readingsBeginUpdate(hash):
    return


async def readingsBulkUpdateIfChanged(hash, reading, value):
    if hash["NAME"] not in readings:
        readings[hash["NAME"]] = {}
    readings[hash["NAME"]][reading] = value


async def readingsBulkUpdate(hash, reading, value, changed=None):
    if hash["NAME"] not in readings:
        readings[hash["NAME"]] = {}
    readings[hash["NAME"]][reading] = value


async def readingsEndUpdate(hash, do_trigger):
    return


async def readingsSingleUpdate(hash, reading, value, do_trigger):
    if hash["NAME"] not in readings:
        readings[hash["NAME"]] = {}
    readings[hash["NAME"]][reading] = value


async def readingsSingleUpdateIfChanged(hash, reading, value, do_trigger):
    if hash["NAME"] not in readings:
        readings[hash["NAME"]] = {}
    readings[hash["NAME"]][reading] = value


async def CommandDefine(hash, definition):
    command_define = definition


async def CommandAttr(hash, attrdef):
    command_attr = attrdef


async def CommandDeleteReading(hash, deldef):
    return


async def checkIfDeviceExists(hash, typeinternal, typevalue, internal, value):
    return False


def convertValue(value):
    if value == None:
        value = ""
    if value == True:
        value = 1
    if value == False:
        value = 0

    return str(value)


async def send_version():
    return



from datetime import datetime
from typing import Optional, Union

DEVICES = {
    # BLE
    152: ["Xiaomi", "Flower Care", "HHCCJCY01"],
    426: ["Xiaomi", "TH Sensor", "LYWSDCGQ/01ZM"],
    839: ["Xiaomi", "Qingping TH Sensor", "CGG1"],
    1034: ["Xiaomi", "Mosquito Repellent", "WX08ZM"],
    1115: ["Xiaomi", "TH Clock", "LYWSD02MMC"],
    1249: ["Xiaomi", "Magic Cube", "XMMF01JQD"],
    1371: ["Xiaomi", "TH Sensor 2", "LYWSD03MMC"],
    1398: ["Xiaomi", "Alarm Clock", "CGD1"],
    1694: ["Aqara", "Door Lock N100", "ZNMS16LM"],
    1747: ["Xiaomi", "ZenMeasure Clock", "MHO-C303"],
    1983: ["Yeelight", "Button S1", "YLAI003"],
    2038: ["Xiaomi", "Night Light 2", "MJYD02YL-A"],
    2443: ["Xiaomi", "Door Sensor 2", "MCCGQ02HL"],
    2480: ["Xiaomi", "Safe Box", "BGX-5/X1-3001"],
    # Mesh
    948: ["Yeelight", "Mesh Downlight", "YLSD01YL"],
    995: ["Yeelight", "Mesh Bulb E14", "YLDP09YL"],
    996: ["Yeelight", "Mesh Bulb E27", "YLDP10YL"],
    997: ["Yeelight", "Mesh Spotlight", "YLSD04YL"],
    1771: ["Xiaomi", "Mesh Bulb", "MJDP09YL"],
    1772: ["Xiaomi", "Mesh Downlight", "MJTS01YL"],
    2076: ["Yeelight", "Mesh Downlight M2", "YLTS02YL/YLTS04YL"],
    2342: ["Yeelight", "Mesh Bulb M2", "YLDP25YL/YLDP26YL"],
    # Mesh Group
    0: ["Xiaomi", "Mesh Group", "Mesh Group"]
}

BLE_FINGERPRINT_ACTION = [
    "Match successful", "Match failed", "Timeout", "Low quality",
    "Insufficient area", "Skin is too dry", "Skin is too wet"
]
BLE_DOOR_ACTION = [
    "Door is open", "Door is closed", "Timeout is not closed",
    "Knock on the door", "Breaking the door", "Door is stuck"
]
BLE_LOCK_ACTION = {
    0b0000: "Unlock outside the door",
    0b0001: "Lock",
    0b0010: "Turn on anti-lock",
    0b0011: "Turn off anti-lock",
    0b0100: "Unlock inside the door",
    0b0101: "Lock inside the door",
    0b0110: "Turn on child lock",
    0b0111: "Turn off child lock",
    0b1111: None
}
BLE_LOCK_METHOD = {
    0b0000: "bluetooth",
    0b0001: "password",
    0b0010: "biological",
    0b0011: "key",
    0b0100: "turntable",
    0b0101: "nfc",
    0b0110: "one-time password",
    0b0111: "two-step verification",
    0b1000: "coercion",
    0b1010: "manual",
    0b1011: "automatic",
    0b1111: None
}
BLE_LOCK_ERROR = {
    0xC0DE0000: "Frequent unlocking with incorrect password",
    0xC0DE0001: "Frequent unlocking with wrong fingerprints",
    0xC0DE0002: "Operation timeout (password input timeout)",
    0xC0DE0003: "Lock picking",
    0xC0DE0004: "Reset button is pressed",
    0xC0DE0005: "The wrong key is frequently unlocked",
    0xC0DE0006: "Foreign body in the keyhole",
    0xC0DE0007: "The key has not been taken out",
    0xC0DE0008: "Error NFC frequently unlocks",
    0xC0DE0009: "Timeout is not locked as required",
    0xC0DE000A: "Failure to unlock frequently in multiple ways",
    0xC0DE000B: "Unlocking the face frequently fails",
    0xC0DE000C: "Failure to unlock the vein frequently",
    0xC0DE000D: "Hijacking alarm",
    0xC0DE000E: "Unlock inside the door after arming",
    0xC0DE000F: "Palmprints frequently fail to unlock",
    0xC0DE0010: "The safe was moved",
    0xC0DE1000: "The battery level is less than 10%",
    0xC0DE1001: "The battery is less than 5%",
    0xC0DE1002: "The fingerprint sensor is abnormal",
    0xC0DE1003: "The accessory battery is low",
    0xC0DE1004: "Mechanical failure",
}

ACTIONS = {
    1249: ['right', 'left'],
    1983: ['single', 'double', 'hold']
}


def get_ble_domain(param: str) -> Optional[str]:
    if param in (
            'sleep', 'lock', 'opening', 'water_leak', 'smoke', 'gas', 'light',
            'contact'):
        return 'binary_sensor'

    elif param in (
            'action', 'rssi', 'temperature', 'humidity', 'illuminance',
            'moisture', 'conductivity', 'battery', 'formaldehyde',
            'mosquitto', 'idle_time'):
        return 'sensor'

    return None


def parse_xiaomi_ble(event: dict, pdid: int) -> Optional[dict]:
    """Parse Xiaomi BLE Data
    https://iot.mi.com/new/doc/embedded-development/ble/object-definition
    """
    eid = event['eid']
    data = bytes.fromhex(event['edata'])
    length = len(data)

    if eid == 0x1001 and length == 3:  # 4097
        if pdid in ACTIONS and data[0] < len(ACTIONS[pdid]):
            return {'action': ACTIONS[pdid][data[0]]}
        else:
            return {'action': data[0]}

    elif eid == 0x1002 and length == 1:  # 4098
        # No sleep (0x00), falling asleep (0x01)
        return {'sleep': data[0]}  # 1 => true

    elif eid == 0x1003 and length == 1:  # 4099
        # Signal strength value
        return {'rssi': data[0]}

    elif eid == 0x1004 and length == 2:  # 4100
        return {'temperature': int.from_bytes(data, 'little', signed=True) / 10.0}

    elif eid == 0x1006 and length == 2:  # 4102
        # Humidity percentage, ranging from 0-1000
        return {'humidity': int.from_bytes(data, 'little') / 10.0}

    elif eid == 0x1007 and length == 3:  # 4103
        # Range: 0-120000, lux
        return {'illuminance': int.from_bytes(data, 'little')}

    elif eid == 0x1008 and length == 1:  # 4104
        # Humidity percentage, range: 0-100
        return {'moisture': data[0]}

    elif eid == 0x1009 and length == 2:  # 4105
        # Soil EC value, Unit us/cm, range: 0-5000
        return {'conductivity': int.from_bytes(data, 'little')}

    elif eid == 0x100A:  # 4106
        # TODO: lock timestamp
        return {'battery': data[0]}

    elif eid == 0x100D and length == 4:  # 4109
        return {
            'temperature': int.from_bytes(data[:2], 'little', signed=True) / 10.0,
            'humidity': int.from_bytes(data[2:], 'little') / 10.0
        }

    elif eid == 0x100E and length == 1:  # 4110
        # 1 => true => on => unlocked
        # 0x00: unlock state (all bolts retracted)
        # TODO: other values
        return {'lock': 1 if data[0] == 0 else 0}

    elif eid == 0x100F and length == 1:  # 4111
        # 1 => true => on => dooor opening
        return {'opening': 1 if data[0] == 0 else 0}

    elif eid == 0x1010 and length == 2:  # 4112
        return {'formaldehyde': int.from_bytes(data, 'little') / 100.0}

    elif eid == 0x1012 and length == 1:  # 4114
        return {'opening': data[0]}  # 1 => true => open

    elif eid == 0x1013 and length == 1:  # 4115
        # Remaining percentage, range 0~100
        return {'mosquitto': data[0]}

    elif eid == 0x1014 and length == 1:  # 4116
        return {'water_leak': data[0]}  # 1 => on => wet

    elif eid == 0x1015 and length == 1:  # 4117
        # TODO: equipment failure (0x02)
        return {'smoke': data[0]}  # 1 => on => alarm

    elif eid == 0x1016 and length == 1:  # 4118
        return {'gas': data[0]}  # 1 => on => alarm

    elif eid == 0x1017 and length == 4:  # 4119
        # The duration of the unmanned state, in seconds
        return {'idle_time': int.from_bytes(data, 'little')}

    elif eid == 0x1018 and length == 1:  # 4120
        return {'light': data[0]}  # 1 => on => strong light

    elif eid == 0x1019 and length == 1:  # 4121
        # 0x00: open the door, 0x01: close the door,
        # 0x02: not closed after timeout, 0x03: device reset
        # 1 => true => open
        if data[0] == 0:
            return {'contact': 1}
        elif data[0] == 1:
            return {'contact': 0}
        else:
            return {}

    elif eid == 0x0006 and len(data) == 5:
        action = int.from_bytes(data[4:], 'little')
        # status, action, state
        return {
            'action': 'fingerprint',
            'action_id': action,
            'key_id': hex(int.from_bytes(data[:4], 'little')),
            'message': BLE_FINGERPRINT_ACTION[action]
        }

    elif eid == 0x0007:
        # TODO: lock timestamp
        return {
            'action': 'door',
            'action_id': data[0],
            'message': BLE_DOOR_ACTION[data[0]]
        }

    elif eid == 0x0008:
        # TODO: lock timestamp
        return {
            'action': 'armed',
            'state': bool(data[0])
        }

    elif eid == 0x000B:  # 11
        action = data[0] & 0x0F
        method = data[0] >> 4
        key_id = int.from_bytes(data[1:5], 'little')
        error = BLE_LOCK_ERROR.get(key_id)

        # all keys except Bluetooth have only 65536 values
        if error is None and method > 0:
            key_id &= 0xFFFF
        elif error:
            key_id = hex(key_id)

        timestamp = int.from_bytes(data[5:], 'little')
        timestamp = datetime.fromtimestamp(timestamp).isoformat()

        return {
            'action': 'lock',
            'action_id': action,
            'method_id': method,
            'message': BLE_LOCK_ACTION[action],
            'method': BLE_LOCK_METHOD[method],
            'key_id': key_id,
            'error': error,
            'timestamp': timestamp
        }

    elif eid == 0x0F:  # 15
        # 1 - moving no light, 100 - moving with light
        # TODO: fix me in future
        return {'action': int.from_bytes(data, 'little')}

    return None


MESH_PROPS = [None, 'light', 'brightness', 'color_temp']


def parse_xiaomi_mesh(data: list):
    """Can receive multiple properties from multiple devices."""
    result = {}

    for payload in data:
        if payload['siid'] != 2 or payload.get('code', 0) != 0:
            continue

        did = payload['did']
        key = MESH_PROPS[payload['piid']]
        result.setdefault(did, {})[key] = payload['value']

    return result


def pack_xiaomi_mesh(did: str, data: Union[dict, list]):
    if isinstance(data, dict):
        return [{
            'did': did,
            'siid': 2,
            'piid': MESH_PROPS.index(k),
            'value': v
        } for k, v in data.items()]
    else:
        return [{
            'did': did,
            'siid': 2,
            'piid': MESH_PROPS.index(k),
        } for k in data]


def get_device(pdid: int, default_name: str) -> Optional[dict]:
    if pdid in DEVICES:
        desc = DEVICES[pdid]
        return {
            'device_manufacturer': desc[0],
            'device_name': desc[0] + ' ' + desc[1],
            'device_model': desc[2] if len(desc) > 2 else pdid
        }
    else:
        return {
            'device_name': default_name,
            'device_model': pdid
        }



import asyncio
import json

from . import utils


class FhemModule:
    def __init__(self, logger):
        self.logger = logger
        self.loop = asyncio.get_event_loop()
        self._tasks = []
        self._conf_set = {}
        self._conf_attr = {}

    def set_attr_config(self, attr_config):
        self._conf_attr = attr_config

    def set_set_config(self, set_config):
        self._conf_set = set_config

    # FHEM FUNCTION
    async def FW_detailFn(self, hash, args, argsh):
        (FW_wname, d, room, pageHash) = args
        ret = """<script type="text/javascript">
        $(document).ready(function() {
          var helpCmdStr = '###SET_CMD_CONFIG_STRING###';
          var helpCmdJson = JSON.parse(helpCmdStr);
          $("select.set").change(helpSetAction);
          function
          helpSetAction(){
            var cmd = $("select.set").val()
            if(helpCmdJson[cmd] && helpCmdJson[cmd].help) {
              if (document.getElementById("idCmdHelp")===null) {
                $('<div id="idCmdHelp" class="makeTable help"></div>')
                      .insertBefore("div.makeTable.internals");
              }
              $("div#idCmdHelp").html(helpCmdJson[cmd].help + "<br>");
            } else {
              $("div#idCmdHelp").remove();
            }
          }
          helpSetAction();

          var helpAttrStr = '###ATTR_CONFIG_STRING###';
          var helpAttrJson = JSON.parse(helpAttrStr);
          $("select.attr").change(helpAttrAction);
          function
          helpAttrAction(){
            var cmd = $("select.attr").val()
            if(helpAttrJson[cmd] && helpAttrJson[cmd].help) {
              if (document.getElementById("idAttrHelp")===null) {
                $('<div id="idAttrHelp" class="makeTable help"></div>')
                      .insertBefore("div.makeTable.attributes");
              }
              $("div#idAttrHelp").html(helpAttrJson[cmd].help + "<br>");
            } else {
              $("div#idAttrHelp").remove();
            }
          }
          helpAttrAction();
        });
        </script>"""
        js_set_conf = {}
        for key in self._conf_set:
            if "help" in self._conf_set[key]:
                js_set_conf[key] = {}
                js_set_conf[key]["help"] = self._conf_set[key]["help"]
        ret = ret.replace(
            "###SET_CMD_CONFIG_STRING###",
            json.dumps(js_set_conf),
        )
        js_attr_conf = {}
        for key in self._conf_attr:
            if "help" in self._conf_attr[key]:
                js_attr_conf[key] = {}
                js_attr_conf[key]["help"] = self._conf_attr[key]["help"]
        ret = ret.replace(
            "###ATTR_CONFIG_STRING###",
            json.dumps(js_attr_conf),
        )
        return ret

    # FHEM FUNCTION
    async def Define(self, hash, args, argsh):
        self.hash = hash
        await utils.handle_define_attr(self._conf_attr, self, hash)

    # FHEM FUNCTION
    async def Attr(self, hash, args, argsh):
        return await utils.handle_attr(self._conf_attr, self, hash, args, argsh)

    # FHEM FUNCTION
    async def Set(self, hash, args, argsh):
        return await utils.handle_set(self._conf_set, self, hash, args, argsh)

    def create_async_task(self, coro):
        task = asyncio.create_task(coro)
        self._tasks.append(task)
        return task

    def cancel_async_task(self, task):
        self._tasks.remove(task)
        task.cancel()

    # FHEM FUNCTION
    async def Undefine(self, hash):
        # cancel all tasks
        for task in self._tasks:
            task.cancel()



import json
import logging
import re
import socket
import time
from telnetlib import Telnet
from threading import Thread
from typing import Optional, Union

from paho.mqtt.client import Client, MQTTMessage
from . import bluetooth, utils
from .mini_miio import SyncmiIO
from .shell import TelnetShell
from .unqlite import Unqlite, SQLite
from .utils import GLOBAL_PROP

_LOGGER = logging.getLogger(__name__)

RE_NWK_KEY = re.compile(r'lumi send-nwk-key (0x.+?) {(.+?)}')
RE_MAC = re.compile('^0x0*')
# MAC reverse
RE_REVERSE = re.compile(r'(..)(..)(..)(..)(..)(..)')


class GatewayV:
    """Handling different firmware versions."""
    ver = None

    @property
    def ver_mesh_group(self) -> str:
        return 'mesh_group_v1' if self.ver >= '1.4.6_0043' else 'mesh_group'

    @property
    def ver_zigbee_db(self) -> str:
        # https://github.com/AlexxIT/XiaomiGateway3/issues/14
        # fw 1.4.6_0012 and below have one zigbee_gw.db file
        # fw 1.4.6_0030 have many json files in this folder
        return '*.json' if self.ver >= '1.4.6_0030' else 'zigbee_gw.db'

    @property
    def ver_miio(self) -> bool:
        return self.ver >= '1.4.7_0063'


class GatewayMesh:
    enabled: bool = None

    devices: dict = None
    updates: dict = None

    miio: SyncmiIO = None
    mesh_params: list = None
    mesh_ts: float = 0

    def debug(self, message: str):
        raise NotImplemented

    def mesh_start(self):
        self.mesh_params = [
            {'did': device['did'], 'siid': 2, 'piid': 1}
            for device in self.devices.values()
            # cannot get state of mesh group
            if device['type'] == 'mesh' and 'childs' not in device
        ]

        if self.mesh_params:
            self.mesh_ts = time.time() + 30
            Thread(target=self.mesh_run).start()

    def mesh_run(self):
        self.debug("Start Mesh Thread")

        while self.enabled:
            if time.time() < self.mesh_ts:
                time.sleep(1)
                continue

            try:
                resp = self.miio.send_bulk('get_properties', self.mesh_params)
                if resp:
                    # get turn on bulbs
                    params = [
                        {'did': item['did'], 'siid': 2, 'piid': 2}
                        for item in resp if item.get('value')
                    ]

                    if params:
                        params += [
                            {'did': item['did'], 'siid': 2, 'piid': 3}
                            for item in params
                        ]
                        resp2 = self.miio.send_bulk('get_properties', params)
                        if resp2:
                            resp += resp2

                    self.debug(f"Pull Mesh {resp}")
                    self.process_mesh_data(resp)

                else:
                    self.debug("Can't get mesh bulb state")

            except Exception as e:
                self.debug(f"ERROR in mesh thread {e}")

            self.mesh_ts = time.time() + 30

    def process_mesh_data(self, data: list):
        data = bluetooth.parse_xiaomi_mesh(data)
        for did, payload in data.items():
            if did in self.updates:
                for handler in self.updates[did]:
                    handler(payload)

    def send_mesh(self, device: dict, data: dict):
        did = device['did']
        payload = bluetooth.pack_xiaomi_mesh(did, data)
        try:
            # 2 seconds are selected experimentally
            if self.miio.send('set_properties', payload):
                self.mesh_force_update()
        except:
            self.debug(f"Can't send mesh {did} => {data}")

    def mesh_force_update(self):
        self.mesh_ts = time.time() + 2


# noinspection PyUnusedLocal
class GatewayStats:
    stats: dict = None
    host: str = None
    info_ts: float = 0
    info_loading: bool = False

    # if mqtt connected
    available: bool = None

    # interval for auto parent refresh in minutes, 0 - disabled auto refresh
    # None - disabled
    parent_scan_interval: Optional[int] = None

    def debug(self, message: str):
        raise NotImplemented

    def add_stats(self, ieee: str, handler):
        self.stats[ieee] = handler

        if self.parent_scan_interval:
            self.info_ts = time.time() + 5

    def remove_stats(self, ieee: str, handler):
        self.stats.pop(ieee)

    def process_gw_stats(self, payload: dict = None):
        # empty payload - update available state
        self.debug(f"gateway <= {payload or self.available}")

        if 'lumi.0' not in self.stats:
            return

        if payload:
            if 'networkUp' in payload:
                # {"networkUp":false}
                payload = {
                    'network_pan_id': payload.get('networkPanId'),
                    'radio_tx_power': payload.get('radioTxPower'),
                    'radio_channel': payload.get('radioChannel'),
                }
            elif 'free_mem' in payload:
                payload.pop('ip')
                payload.pop('ssid')
                s = payload.pop('run_time')
                h, m, s = s // 3600, s % 3600 // 60, s % 60
                payload['uptime'] = f"{h:02}:{m:02}:{s:02}"

        self.stats['lumi.0'](payload)

    def process_zb_stats(self, payload: dict):
        ieee = payload['eui64']
        if ieee in self.stats:
            self.stats[ieee](payload)

        if self.info_ts and time.time() > self.info_ts:
            self.get_gateway_info()

    def process_ble_stats(self, payload: dict):
        did = payload['dev']['did']
        if did in self.stats:
            self.stats[did](payload)

    def get_gateway_info(self):
        if self.info_loading:
            return
        self.info_loading = True
        Thread(target=self._info_loader_run).start()

    def _info_loader_run(self):
        self.debug("Update parent info table")

        telnet = Telnet(self.host, 4901)
        telnet.read_until(b'Lumi_Z3GatewayHost')

        telnet.write(b"option print-rx-msgs disable\r\n")
        telnet.read_until(b'Lumi_Z3GatewayHost')

        telnet.write(b"plugin device-table print\r\n")
        raw = telnet.read_until(b'Lumi_Z3GatewayHost').decode()
        m1 = re.findall(r'\d+ ([A-F0-9]{4}): {2}([A-F0-9]{16}) 0 {2}\w+ (\d+)',
                        raw)

        telnet.write(b"plugin stack-diagnostics child-table\r\n")
        raw = telnet.read_until(b'Lumi_Z3GatewayHost').decode()
        m2 = re.findall(r'\(>\)([A-F0-9]{16})', raw)

        telnet.write(b"plugin stack-diagnostics neighbor-table\r\n")
        raw = telnet.read_until(b'Lumi_Z3GatewayHost').decode()
        m3 = re.findall(r'\(>\)([A-F0-9]{16})', raw)

        telnet.write(b"plugin concentrator print-table\r\n")
        raw = telnet.read_until(b'Lumi_Z3GatewayHost').decode()
        m4 = re.findall(r': (.{16,}) -> 0x0000', raw)
        m4 = [i.replace('0x', '').split(' -> ') for i in m4]
        m4 = {i[0]: i[1:] for i in m4}

        for i in m1:
            ieee = '0x' + i[1]
            if ieee not in self.stats:
                continue

            nwk = i[0]
            ago = int(i[2])
            type_ = 'device' if i[1] in m2 else 'router' if i[1] in m3 else '-'
            parent = '0x' + m4[nwk][0] if nwk in m4 else '-'

            self.stats[ieee]({
                'nwk': '0x' + nwk,
                'ago': ago,
                'type': type_,
                'parent': parent
            })

        self.info_loading = False
        # one hour later
        if self.parent_scan_interval:
            self.info_ts = time.time() + self.parent_scan_interval * 60


# noinspection PyUnusedLocal
class Gateway3(Thread, GatewayV, GatewayMesh, GatewayStats):
    pair_model = None
    pair_payload = None

    def __init__(self, host: str, token: str, config: dict, **options):
        super().__init__(daemon=True)

        self.host = host
        self.options = options

        self.miio = SyncmiIO(host, token)

        self.mqtt = Client()
        self.mqtt.on_connect = self.on_connect
        self.mqtt.on_disconnect = self.on_disconnect
        self.mqtt.on_message = self.on_message
        self.mqtt.connect_async(host)

        self._ble = options.get('ble')  # for fast access
        self._debug = options.get('debug', '')  # for fast access
        self.parent_scan_interval = options.get('parent')  # for fast access
        self.default_devices = config['devices']

        self.devices = {}
        self.updates = {}
        self.setups = {}
        self.stats = {}

    @property
    def device(self):
        return self.devices['lumi.0']

    def add_update(self, did: str, handler):
        """Add handler to device update event."""
        self.updates.setdefault(did, []).append(handler)

    def remove_update(self, did: str, handler):
        self.updates.setdefault(did, []).remove(handler)

    def add_setup(self, domain: str, handler):
        """Add hass device setup funcion."""
        self.setups[domain] = handler

    def debug(self, message: str):
        _LOGGER.debug(f"{self.host} | {message}")

    def stop(self):
        self.enabled = False
        self.mqtt.loop_stop()

    def run(self):
        """Main thread loop."""
        self.enabled = True
        while self.enabled:
            # if not telnet - enable it
            if not self._check_port(23) and not self._enable_telnet():
                time.sleep(30)
                continue

            devices = self._prepeare_gateway(with_devices=True)
            if devices:
                self.setup_devices(devices)
                break

        self.mesh_start()

        while self.enabled:
            # if not telnet - enable it
            if not self._check_port(23) and not self._enable_telnet():
                time.sleep(30)
                continue

            # if not mqtt - enable it (handle Mi Home and ZHA mode)
            if not self._mqtt_connect() and not self._prepeare_gateway():
                time.sleep(60)
                continue

            self.mqtt.loop_forever()

    def _check_port(self, port: int):
        """Check if gateway port open."""
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        try:
            return s.connect_ex((self.host, port)) == 0
        finally:
            s.close()

    def _enable_telnet(self):
        """Enable telnet with miio protocol."""
        if self.miio.send("enable_telnet_service") != 'ok':
            self.debug(f"Can't enable telnet")
            return False
        return True

    def _prepeare_gateway(self, with_devices: bool = False):
        """Launching the required utilities on the hub, if they are not already
        running.
        """
        self.debug("Prepare Gateway")
        try:
            shell = TelnetShell(self.host)

            self.ver = shell.get_version()
            self.debug(f"Version: {self.ver}")

            ps = shell.get_running_ps()

            if "mosquitto -d" not in ps:
                self.debug("Run public mosquitto")
                shell.run_public_mosquitto()

            # all data or only necessary events
            pattern = '\\{"' if 'miio' in self._debug \
                else "ble_event|properties_changed|heartbeat"

            if f"awk /{pattern} {{" not in ps:
                self.debug(f"Redirect miio to MQTT")
                shell.redirect_miio2mqtt(pattern, self.ver_miio)

            if self.options.get('buzzer'):
                if "basic_gw -b" in ps:
                    self.debug("Disable buzzer")
                    shell.stop_buzzer()
            else:
                if "dummy:basic_gw" in ps:
                    self.debug("Enable buzzer")
                    shell.run_buzzer()

            if self.options.get('zha'):
                if "socat" not in ps:
                    if "Received" in shell.check_or_download_socat():
                        self.debug("Download socat")
                    self.debug("Run socat")
                    shell.run_socat()

                if "Lumi_Z3GatewayHost_MQTT" in ps:
                    self.debug("Stop Lumi Zigbee")
                    shell.stop_lumi_zigbee()

            else:
                if "socat" in ps:
                    self.debug("Stop socat")
                    shell.stop_socat()

                if (self.parent_scan_interval is not None and
                        "Lumi_Z3GatewayHost_MQTT -n 1 -b 115200 -v" not in ps):
                    self.debug("Run public Zigbee console")
                    shell.run_public_zb_console()

                elif "Lumi_Z3GatewayHost_MQTT" not in ps:
                    self.debug("Run Lumi Zigbee")
                    shell.run_lumi_zigbee()

            if with_devices:
                self.debug("Get devices")
                return self._get_devices(shell)

            return True

        except (ConnectionRefusedError, socket.timeout):
            return False

        except Exception as e:
            self.debug(f"Can't read devices: {e}")
            return False

    def _mqtt_connect(self) -> bool:
        try:
            self.mqtt.reconnect()
            return True
        except:
            return False

    def _miio_connect(self) -> bool:
        if not self.miio.ping():
            self.debug("Can't send handshake")
            return False

        return True

    def _get_devices(self, shell: TelnetShell):
        """Load devices info for Coordinator, Zigbee and Mesh."""

        # 1. Read coordinator info
        raw = shell.read_file('/data/zigbee/coordinator.info')
        device = json.loads(raw)
        devices = [{
            'did': 'lumi.0',
            'model': 'lumi.gateway.mgl03',
            'mac': device['mac'],
            'type': 'gateway',
            'init': {
                'firmware lock': shell.check_firmware_lock()
            }
        }]

        # 2. Read zigbee devices
        if not self.options.get('zha'):
            raw = shell.read_file('/data/zigbee_gw/' + self.ver_zigbee_db,
                                  as_base64=True)
            if raw.startswith(b'unqlite'):
                db = Unqlite(raw)
                data = db.read_all()
            else:
                raw = re.sub(br'}\s+{', b',', raw)
                data = json.loads(raw)

            # data = {} or data = {'dev_list': 'null'}
            dev_list = json.loads(data.get('dev_list', 'null')) or []

            for did in dev_list:
                model = data[did + '.model']
                desc = utils.get_device(model)

                # skip unknown model
                if desc is None:
                    self.debug(f"{did} has an unsupported modell: {model}")
                    continue

                retain = json.loads(data[did + '.prop'])['props']
                self.debug(f"{did} {model} retain: {retain}")

                params = {
                    p[2]: retain.get(p[1])
                    for p in (desc['params'] or desc['mi_spec'])
                    if p[1] is not None
                }

                device = {
                    'did': did,
                    'mac': '0x' + data[did + '.mac'],
                    'model': data[did + '.model'],
                    'type': 'zigbee',
                    'zb_ver': data[did + '.version'],
                    'init': utils.fix_xiaomi_props(params),
                    'online': retain.get('alive', 1) == 1
                }
                devices.append(device)

        # 3. Read bluetooth devices
        if self._ble:
            raw = shell.read_file('/data/miio/mible_local.db', as_base64=True)
            db = SQLite(raw)

            # load BLE devices
            rows = db.read_table('gateway_authed_table')
            for row in rows:
                device = {
                    'did': row[4],
                    'mac': RE_REVERSE.sub(r'\6\5\4\3\2\1', row[1]),
                    'model': row[2],
                    'type': 'ble'
                }
                devices.append(device)

            # load Mesh groups
            try:
                mesh_groups = {}

                rows = db.read_table(self.ver_mesh_group)
                for row in rows:
                    # don't know if 8 bytes enougth
                    mac = int(row[0]).to_bytes(8, 'big').hex()
                    device = {
                        'did': 'group.' + row[0],
                        'mac': mac,
                        'model': 0,
                        'childs': [],
                        'type': 'mesh'
                    }
                    group_addr = row[1]
                    mesh_groups[group_addr] = device

                # load Mesh bulbs
                rows = db.read_table('mesh_device')
                for row in rows:
                    device = {
                        'did': row[0],
                        'mac': row[1].replace(':', ''),
                        'model': row[2],
                        'type': 'mesh'
                    }
                    devices.append(device)

                    group_addr = row[5]
                    if group_addr in mesh_groups:
                        # add bulb to group if exist
                        mesh_groups[group_addr]['childs'].append(row[0])

                for device in mesh_groups.values():
                    if device['childs']:
                        devices.append(device)

            except:
                _LOGGER.exception("Can't read mesh devices")

        # for testing purposes
        for k, v in self.default_devices.items():
            if k[0] == '_':
                devices.append(v)

        return devices

    def lock_firmware(self, enable: bool):
        self.debug(f"Set firmware lock to {enable}")
        try:
            shell = TelnetShell(self.host)
            if "Received" in shell.check_or_download_busybox():
                self.debug("Download busybox")
            shell.lock_firmware(enable)
            locked = shell.check_firmware_lock()
            shell.close()
            return enable == locked

        except Exception as e:
            self.debug(f"Can't set firmware lock: {e}")
            return False

    def on_connect(self, client, userdata, flags, rc):
        self.debug("MQTT connected")
        self.mqtt.subscribe('#')

        self.available = True
        self.process_gw_stats()

    def on_disconnect(self, client, userdata, rc):
        self.debug("MQTT disconnected")
        # force end mqtt.loop_forever()
        self.mqtt.disconnect()

        self.available = False
        self.process_gw_stats()

    def on_message(self, client: Client, userdata, msg: MQTTMessage):
        if 'mqtt' in self._debug:
            self.debug(f"[MQ] {msg.topic} {msg.payload.decode()}")

        if msg.topic == 'zigbee/send':
            payload = json.loads(msg.payload)
            self.process_message(payload)

        elif msg.topic == 'log/miio':
            if 'miio' in self._debug:
                self.debug(f"[MI] {msg.payload}")

            if self._ble and (
                    b'_async.ble_event' in msg.payload or
                    b'properties_changed' in msg.payload or
                    b'event.gw.heartbeat' in msg.payload
            ):
                try:
                    for raw in utils.extract_jsons(msg.payload):
                        if b'_async.ble_event' in raw:
                            data = json.loads(raw)['params']
                            self.process_ble_event(data)
                            self.process_ble_stats(data)
                        elif b'properties_changed' in raw:
                            data = json.loads(raw)['params']
                            self.debug(f"Process props {data}")
                            self.process_mesh_data(data)
                        elif b'event.gw.heartbeat' in raw:
                            payload = json.loads(raw)['params'][0]
                            self.process_gw_stats(payload)
                except:
                    _LOGGER.warning(f"Can't read BT: {msg.payload}")

        elif msg.topic.endswith('/heartbeat'):
            payload = json.loads(msg.payload)
            self.process_gw_stats(payload)

        elif msg.topic.endswith(('/MessageReceived', '/devicestatechange')):
            payload = json.loads(msg.payload)
            self.process_zb_stats(payload)

        # read only retained ble
        elif msg.topic.startswith('ble') and msg.retain:
            payload = json.loads(msg.payload)
            self.process_ble_retain(msg.topic[4:], payload)

        elif self.pair_model and msg.topic.endswith('/commands'):
            self.process_pair(msg.payload)

    def setup_devices(self, devices: list):
        """Add devices to hass."""
        for device in devices:
            if device['type'] in ('gateway', 'zigbee'):
                desc = utils.get_device(device['model'])
                if not desc:
                    self.debug(f"Unsupported model: {device}")
                    continue

                self.debug(f"Setup Zigbee device {device}")

                device.update(desc)

                # update params from config
                default_config = (
                        self.default_devices.get(device['mac']) or
                        self.default_devices.get(device['did'])
                )
                if default_config:
                    device.update(default_config)

                self.devices[device['did']] = device

                for param in (device['params'] or device['mi_spec']):
                    domain = param[3]
                    if not domain:
                        continue

                    # wait domain init
                    while domain not in self.setups:
                        time.sleep(1)

                    attr = param[2]
                    self.setups[domain](self, device, attr)

            elif device['type'] == 'mesh':
                desc = bluetooth.get_device(device['model'], 'Mesh')
                device.update(desc)

                self.debug(f"Setup Mesh device {device}")

                # update params from config
                default_config = self.default_devices.get(device['did'])
                if default_config:
                    device.update(default_config)

                device['online'] = False

                self.devices[device['did']] = device

                # wait domain init
                while 'light' not in self.setups:
                    time.sleep(1)

                self.setups['light'](self, device, 'light')

            elif device['type'] == 'ble':
                # only save info for future
                desc = bluetooth.get_device(device['model'], 'BLE')
                device.update(desc)

                # update params from config
                default_config = self.default_devices.get(device['did'])
                if default_config:
                    device.update(default_config)

                self.devices[device['did']] = device

                device['init'] = {}

            if self.options.get('stats') and device['type'] != 'mesh':
                while 'sensor' not in self.setups:
                    time.sleep(1)
                self.setups['sensor'](self, device, device['type'])

    def process_message(self, data: dict):
        if data['cmd'] == 'heartbeat':
            # don't know if only one item
            assert len(data['params']) == 1, data

            data = data['params'][0]
            pkey = 'res_list'
        elif data['cmd'] == 'report':
            pkey = 'params' if 'params' in data else 'mi_spec'
        elif data['cmd'] in ('write_rsp', 'read_rsp'):
            pkey = 'results'
        elif data['cmd'] == 'write_ack':
            return
        else:
            _LOGGER.warning(f"Unsupported cmd: {data}")
            return

        did = data['did']

        # skip without callback
        if did not in self.updates:
            return

        device = self.devices[did]
        payload = {}

        # convert codes to names
        for param in data[pkey]:
            if param.get('error_code', 0) != 0:
                continue

            prop = param['res_name'] if 'res_name' in param else \
                f"{param['siid']}.{param['piid']}"

            if prop in GLOBAL_PROP:
                prop = GLOBAL_PROP[prop]
            else:
                prop = next((
                    p[2] for p in (device['params'] or device['mi_spec'])
                    if p[0] == prop
                ), prop)

            if prop in ('temperature', 'humidity', 'pressure'):
                payload[prop] = param['value'] / 100.0
            elif prop == 'battery' and param['value'] > 1000:
                # xiaomi light sensor
                payload[prop] = round((min(param['value'], 3200) - 2500) / 7)
            elif prop == 'alive':
                # {'res_name':'8.0.2102','value':{'status':'online','time':0}}
                device['online'] = (param['value']['status'] == 'online')
            elif prop == 'angle':
                # xiaomi cube 100 points = 360 degrees
                payload[prop] = param['value'] * 4
            elif prop == 'duration':
                # xiaomi cube
                payload[prop] = param['value'] / 1000.0
            elif prop in ('consumption', 'power'):
                payload[prop] = round(param['value'], 2)
            else:
                payload[prop] = param['value']

        self.debug(f"{device['did']} {device['model']} <= {payload}")

        for handler in self.updates[did]:
            handler(payload)

        if 'added_device' in payload:
            # {'did': 'lumi.fff', 'mac': 'fff', 'model': 'lumi.sen_ill.mgl01',
            # 'version': '21', 'zb_ver': '3.0'}
            device = payload['added_device']
            device['mac'] = '0x' + device['mac']
            device['type'] = 'zigbee'
            device['init'] = payload
            self.setup_devices([device])

    def process_ble_event(self, data: dict):
        self.debug(f"Process BLE {data}")

        pdid = data['dev'].get('pdid')

        did = data['dev']['did']
        if did not in self.devices:
            mac = data['dev']['mac'].replace(':', '').lower() \
                if 'mac' in data['dev'] else \
                'ble_' + did.replace('blt.3.', '')
            self.devices[did] = device = {
                'did': did, 'mac': mac, 'init': {}, 'type': 'bluetooth'}
            desc = bluetooth.get_device(pdid, 'BLE')
            device.update(desc)

            # update params from config
            default_config = self.default_devices.get(did)
            if default_config:
                device.update(default_config)

        else:
            device = self.devices[did]

        if isinstance(data['evt'], list):
            # check if only one
            assert len(data['evt']) == 1, data
            payload = bluetooth.parse_xiaomi_ble(data['evt'][0], pdid)
        elif isinstance(data['evt'], dict):
            payload = bluetooth.parse_xiaomi_ble(data['evt'], pdid)
        else:
            payload = None

        if payload is None:
            self.debug(f"Unsupported BLE {data}")
            return

        # init entities if needed
        init = device['init']
        for k in payload.keys():
            if k in init:
                # update for retain
                init[k] = payload[k]
                continue

            init[k] = payload[k]

            domain = bluetooth.get_ble_domain(k)
            if not domain:
                continue

            # wait domain init
            while domain not in self.setups:
                time.sleep(1)

            self.setups[domain](self, device, k)

        if did in self.updates:
            for handler in self.updates[did]:
                handler(payload)

        raw = json.dumps(init, separators=(',', ':'))
        self.mqtt.publish(f"ble/{did}", raw, retain=True)

    def process_ble_retain(self, did: str, payload: dict):
        if did not in self.devices:
            self.debug(f"BLE device {did} is no longer on the gateway")
            return

        self.debug(f"{did} retain: {payload}")

        device = self.devices[did]

        # init entities if needed
        for k in payload.keys():
            # don't retain action
            if k in device['init'] or k == 'action':
                continue

            device['init'][k] = payload[k]

            domain = bluetooth.get_ble_domain(k)
            if not domain:
                continue

            # wait domain init
            while domain not in self.setups:
                time.sleep(1)

            self.setups[domain](self, device, k)

        if did in self.updates:
            for handler in self.updates[did]:
                handler(payload)

    def process_pair(self, raw: bytes):
        # get shortID and eui64 of paired device
        if b'lumi send-nwk-key' in raw:
            # create model response
            payload = f"0x18010105000042{len(self.pair_model):02x}" \
                      f"{self.pair_model.encode().hex()}"
            m = RE_NWK_KEY.search(raw.decode())
            self.pair_payload = json.dumps({
                'sourceAddress': m[1],
                'eui64': '0x' + m[2],
                'profileId': '0x0104',
                'clusterId': '0x0000',
                'sourceEndpoint': '0x01',
                'destinationEndpoint': '0x01',
                'APSCounter': '0x01',
                'APSPlayload': payload
            }, separators=(',', ':'))

        # send model response "from device"
        elif b'zdo active ' in raw:
            mac = self.device['mac'][2:].upper()
            self.mqtt.publish(f"gw/{mac}/MessageReceived", self.pair_payload)

    def send(self, device: dict, data: dict):
        payload = {'cmd': 'write', 'did': device['did']}

        # convert hass prop to lumi prop
        if device['mi_spec']:
            params = []
            for k, v in data.items():
                if k == 'switch':
                    v = bool(v)
                k = next(p[0] for p in device['mi_spec'] if p[2] == k)
                params.append({'siid': k[0], 'piid': k[1], 'value': v})

            payload['mi_spec'] = params
        else:
            params = [{
                'res_name': next(p[0] for p in device['params'] if p[2] == k),
                'value': v
            } for k, v in data.items()]

            payload = {
                'cmd': 'write',
                'did': device['did'],
                'params': params,
            }

        self.debug(f"{device['did']} {device['model']} => {payload}")
        payload = json.dumps(payload, separators=(',', ':')).encode()
        self.mqtt.publish('zigbee/recv', payload)

    def send_telnet(self, *args: str):
        try:
            shell = TelnetShell(self.host)
            for command in args:
                if command == 'ftp':
                    shell.check_or_download_busybox()
                    shell.run_ftp()
                else:
                    shell.exec(command)
            shell.close()

        except Exception as e:
            _LOGGER.exception(f"Telnet command error: {e}")

    def send_mqtt(self, cmd: str):
        if cmd == 'publishstate':
            mac = self.device['mac'][2:].upper()
            self.mqtt.publish(f"gw/{mac}/publishstate")

    def get_device(self, mac: str) -> Optional[dict]:
        for device in self.devices.values():
            if device.get('mac') == mac:
                return device
        return None


def is_gw3(host: str, token: str) -> Optional[str]:
    device = SyncmiIO(host, token)
    info = device.info()

    if not info:
        return 'cant_connect'

    if info['model'] != 'lumi.gateway.mgl03':
        return 'wrong_model'

    return None



import asyncio
import aiohttp

from .. import fhem
from .. import utils
from fhempy.lib.generic import FhemModule

BASE_URL = "http://www.wienerlinien.at/ogd_realtime/monitor?rbl={}"

DEPARTURES = {
    "first": {"key": 0, "name": "{} first departure"},
    "next": {"key": 1, "name": "{} next departure"},
}


class wienerlinien(FhemModule):
    def __init__(self, logger):
        super().__init__(logger)
        self.firstnext = "first"
        self._updateloop = None
        self._last_data = None
        return

    # FHEM FUNCTION
    async def Define(self, hash, args, argsh):
        await super().Define(hash, args, argsh)
        set_list_conf = {"update": {}}
        self.set_set_config(set_list_conf)
        if len(args) < 4:
            return "Usage: define devname PythonModule wienerlinien <STOPID>"
        self._stopid = args[3]
        self.api = WienerlinienAPI(self._stopid)
        self._updateloop = self.create_async_task(self.update_loop())
        # delete all readings on define
        self.create_async_task(fhem.CommandDeleteReading(hash, hash["NAME"] + " .*"))

    async def set_update(self, hash, params):
        self.create_async_task(self.update())
        return ""

    async def update_loop(self):
        while True:
            await self.update()
            await asyncio.sleep(30)

    async def update(self):
        try:
            data = await self.api.get_json()
            self.logger.debug(data)
            if data is None:
                return
            message = data.get("message", {})
            data = data.get("data", {})
        except:
            self.logger.debug("Could not get new state")
            return

        if data is None:
            return
        try:
            if len(data["monitors"]) == 0:
                flat_data = {}
                flat_data_location = {}
                state_text = "no departures"
            else:
                flat_data = utils.flatten_json(data["monitors"][0]["lines"][0])
                flat_data_location = utils.flatten_json(
                    data["monitors"][0]["locationStop"]
                )
                state_text = (
                    flat_data["towards"]
                    + ": "
                    + str(flat_data["departures_departure_0_departureTime_countdown"])
                )

            if self._last_data:
                del_readings = set(self._last_data) - set(flat_data)
            else:
                del_readings = {}

            await fhem.readingsBeginUpdate(self.hash)
            for msg in message:
                await fhem.readingsBulkUpdateIfChanged(
                    self.hash, "msg_" + msg, message[msg]
                )
            for data_name in flat_data:
                await fhem.readingsBulkUpdateIfChanged(
                    self.hash, "line_" + data_name, flat_data[data_name]
                )
            for data_name in flat_data_location:
                await fhem.readingsBulkUpdateIfChanged(
                    self.hash, "loc_" + data_name, flat_data_location[data_name]
                )

            if "trafficjam" in flat_data and flat_data["trafficjam"] == 1:
                state_text += " (traffic jam)"
            await fhem.readingsBulkUpdateIfChanged(self.hash, "state", state_text)
            await fhem.readingsEndUpdate(self.hash, 1)

            # delete old readings which were not updated
            for del_reading in del_readings:
                await fhem.CommandDeleteReading(
                    self.hash, self.hash["NAME"] + " line_" + del_reading
                )

            self._last_data = flat_data

        except Exception:
            self.logger.exception("Failed...")
            pass


class WienerlinienAPI:
    """Call API."""

    def __init__(self, stopid):
        """Initialize."""
        self.session = aiohttp.ClientSession()
        self.stopid = stopid

    async def get_json(self):
        """Get json from API endpoint."""
        value = None
        url = BASE_URL.format(self.stopid)
        try:
            response = await self.session.get(url)
            value = await response.json()
        except Exception:
            self.logger.exception("Failed...")
            pass

        return value



import asyncio
import concurrent
import dbus
import functools
import re
import subprocess

from .. import fhem, utils
from fhempy.lib.generic import FhemModule
import datetime


class ble_reset(FhemModule):
    def __init__(self, logger):
        super().__init__(logger)
        self._hours = 24
        self._resettask = None
        self._attr_list = {"reset_time": {"default": "04:00", "format": "str"}}
        self.set_attr_config(self._attr_list)
        set_list_conf = {
            "interval": {"args": ["hours"], "options": "1h,2h,4h,8h,12h,24h,manual"},
            "resetnow": {},
        }
        self.set_set_config(set_list_conf)
        return

    def get_hci_ifaces(self):
        iface_list = []
        bus = dbus.SystemBus()
        manager = dbus.Interface(
            bus.get_object("org.bluez", "/"), "org.freedesktop.DBus.ObjectManager"
        )
        objects = manager.GetManagedObjects()
        for path, interfaces in objects.items():
            adapter = interfaces.get("org.bluez.Adapter1")
            if adapter is None:
                continue
            iface_list.append(re.search(r"\d+$", path)[0])
        return iface_list

    # FHEM FUNCTION
    async def Define(self, hash, args, argsh):
        await super().Define(hash, args, argsh)
        self._reset_time = datetime.datetime.strptime(self._attr_reset_time, "%H:%M")

        hours = await fhem.ReadingsVal(hash["NAME"], "interval", "24h")
        if hours == "manual":
            self._hours = 0
        else:
            self._hours = int(hours[:-1])
            self._resettask = self.create_async_task(self.ble_reset())

        await fhem.readingsBeginUpdate(hash)
        await fhem.readingsBulkUpdateIfChanged(hash, "interval", hours)
        await fhem.readingsBulkUpdateIfChanged(hash, "state", "active")
        await fhem.readingsEndUpdate(hash, 1)

    async def ble_reset(self):
        while True:
            if self._hours > 0:
                now = datetime.datetime.now()
                first_reset = datetime.datetime(
                    now.year,
                    now.month,
                    now.day,
                    self._reset_time.hour,
                    self._reset_time.minute,
                )
                # calculate next reset time
                if (now - first_reset).total_seconds() > 0:
                    next_reset = now + datetime.timedelta(
                        seconds=(self._hours * 3600)
                        - ((now - first_reset).seconds % (self._hours * 3600))
                    )
                else:
                    next_reset = now + datetime.timedelta(
                        seconds=1 + ((first_reset - now).seconds % (self._hours * 3600))
                    )
                await fhem.readingsSingleUpdateIfChanged(
                    self.hash,
                    "nextreset",
                    f"{next_reset.hour:02}:{next_reset.minute:02}",
                    1,
                )
                await asyncio.sleep((next_reset - now).seconds)
            # do reset now
            await self.ble_reset_once()

            now = datetime.datetime.now()
            await fhem.readingsSingleUpdate(
                self.hash, "lastreset", f"{now.hour:02}:{now.minute:02}", 1
            )
            if self._hours == 0:
                return

    async def ble_reset_once(self):
        with concurrent.futures.ThreadPoolExecutor() as pool:
            await asyncio.get_event_loop().run_in_executor(
                pool, functools.partial(self.do_ble_reset)
            )
        now = datetime.datetime.now()
        await fhem.readingsSingleUpdate(
            self.hash, "lastreset", f"{now.hour:02}:{now.minute:02}", 1
        )

    def do_ble_reset(self):
        try:
            ifaces = self.get_hci_ifaces()
            subprocess.Popen(["sudo", "systemctl", "restart", "bluetooth"]).wait()
            for iface in ifaces:
                subprocess.Popen(["sudo", "hciconfig", "hci" + iface, "reset"]).wait()
                subprocess.Popen(["sudo", "hciconfig", "hci" + iface, "up"]).wait()
        except:
            self.logger.exception("Failed to reset bluetooth")

    async def set_interval(self, hash, params):
        if self._resettask:
            self._resettask.cancel()

        hours = params["hours"]
        if hours == "manual":
            self._hours = 0
            await fhem.readingsSingleUpdate(hash, "nextreset", "-", 1)
        else:
            self._hours = int(hours[:-1])
        await fhem.readingsSingleUpdate(hash, "interval", hours, 1)

        self._resettask = self.create_async_task(self.ble_reset())

    async def set_resetnow(self, hash, params):
        self.create_async_task(self.ble_reset_once())

    async def set_attr_reset_time(self, hash):
        if self._resettask:
            self._resettask.cancel()
        self._reset_time = datetime.datetime.strptime(self._attr_reset_time, "%H:%M")
        self._resettask = self.create_async_task(self.ble_reset())



import asyncio

from .. import fhem
from .. import utils
from ..generic import FhemModule


class helloworld(FhemModule):
    def __init__(self, logger):
        super().__init__(logger)

        attr_config = {
            "interval": {
                "default": 100,
                "format": "int",
                "help": "Change interval, default is 100.",
            }
        }
        self.set_attr_config(attr_config)

        set_config = {
            "mode": {
                "args": ["mode"],
                "argsh": ["mode"],
                "params": {"mode": {"default": "eco", "optional": False}},
                "options": "eco,comfort",
            },
            "desiredTemp": {"args": ["temperature"], "options": "slider,10,1,30"},
            "holidayMode": {
                "args": ["endday", "endtime", "temperature"],
                "params": {
                    "endday": {"default": "31.12.2030"},
                    "endtime": {"default": "23:59"},
                    "temperature": {"default": 21, "format": "int"},
                },
            },
            "on": {
                "args": ["seconds"],
                "params": {
                    "seconds": {"default": 0, "optional": True, "format": "int"}
                },
                "help": "Specify seconds as parameter to change to off after X seconds.",
            },
            "off": {},
        }
        self.set_set_config(set_config)

    # FHEM FUNCTION
    async def Define(self, hash, args, argsh):
        await super().Define(hash, args, argsh)
        if len(args) > 3:
            return "Usage: define hello_fhempy PythonModule helloworld"
        await fhem.readingsBeginUpdate(hash)
        await fhem.readingsBulkUpdateIfChanged(hash, "state", "on")
        await fhem.readingsEndUpdate(hash, 1)

    # Attribute function format: set_attr_NAMEOFATTRIBUTE(self, hash)
    # self._attr_NAMEOFATTRIBUTE contains the new state
    async def set_attr_interval(self, hash):
        # attribute was set to self._attr_interval
        # you can use self._attr_interval already with the new variable
        pass

    # Set functions in format: set_NAMEOFSETFUNCTION(self, hash, params)
    async def set_on(self, hash, params):
        # params contains the keyword which was defined in set_list_conf for "on"
        # if not provided by the user it will be "" as defined in set_list_conf (default = "" and optional = True)
        seconds = params["seconds"]
        if seconds != 0:
            await fhem.readingsSingleUpdate(hash, "state", "on " + str(seconds), 1)
        else:
            await fhem.readingsSingleUpdate(hash, "state", "on", 1)

    async def set_off(self, hash, params):
        # no params argument here, as set_off doesn't have arguments defined in set_list_conf
        await fhem.readingsSingleUpdate(hash, "state", "off", 1)
        self.create_async_task(self.long_running_task())
        return ""

    async def long_running_task(self):
        await asyncio.sleep(30)
        await fhem.readingsSingleUpdate(self.hash, "state", "long running off", 1)

    async def set_mode(self, hash, params):
        # user can specify mode as mode=eco or just eco as argument
        # params['mode'] contains the mode provided by user
        mode = params["mode"]
        await fhem.readingsSingleUpdate(hash, "mode", mode, 1)

    async def set_desiredTemp(self, hash, params):
        temp = params["temperature"]
        await fhem.readingsSingleUpdate(hash, "mode", temp, 1)

    async def set_holidayMode(self, hash, params):
        start = params["start"]
        end = params["end"]
        temp = params["temperature"]
        await fhem.readingsSingleUpdate(hash, "start", start, 1)
        await fhem.readingsSingleUpdate(hash, "end", end, 1)
        await fhem.readingsSingleUpdate(hash, "temp", temp, 1)



import asyncio
import logging
import pytest
import os
from fhempy.lib.pkg_installer import check_and_install_dependencies
from tests.utils import mock_fhem

import requests_mock


def load_fixture(filename):
    """Load a fixture."""
    path = os.path.join(os.path.dirname(__file__), "fixtures", filename)
    with open(path) as fdp:
        return fdp.read()


@pytest.fixture(autouse=True)
def mock_tuya_requests():

    with requests_mock.Mocker() as mock:
        mock.get(
            "https://openapi.tuyaeu.com/v1.0/token?grant_type=1",
            text=load_fixture("token.json"),
        )
        mock.get(
            "https://openapi.tuyaeu.com/v1.0/devices/345678345673456567",
            text=load_fixture("device.json"),
        )
        mock.get(
            "https://openapi.tuyaeu.com/v1.0/users/12345678901234567890/devices",
            text=load_fixture("devices.json"),
        )
        yield mock


@pytest.mark.asyncio
async def test_mapping_ctrl(mocker):
    # prepare
    mock_fhem.mock_module(mocker)

    testval = None
    testindex = None

    def set_status(instance, value, switch):
        nonlocal testval, testindex
        testval = value
        testindex = switch

    def set_value(instance, index, value):
        nonlocal testval, testindex
        testval = value
        testindex = index

    def status(instance):
        return {
            "dps": {
                "1": False,
                "9": 0,
                "17": 2,
                "18": 0,
                "19": 0,
                "20": 2371,
                "21": 1,
                "22": 691,
                "23": 30757,
                "24": 18008,
                "25": 2320,
                "26": 0,
                "38": "off",
                "41": "",
                "42": "",
                "46": False,
            }
        }

    mocker.patch("tinytuya.Device.set_status", set_status)
    mocker.patch("tinytuya.Device.set_value", set_value)
    mocker.patch("tinytuya.Device.status", status)

    testhash = {"NAME": "testdevice"}
    await check_and_install_dependencies("tuya")
    from fhempy.lib.tuya.tuya import tuya

    fhempy_device = tuya(logging.getLogger(__name__))

    await fhempy_device.Define(
        testhash,
        [
            "testdevice",
            "PythonModule",
            "tuya",
            "j0zozzoarutv0nu1",
            "34567adf13165",
            "192.168.86.35",
            "345678345678",
            "3.3",
            "12345678123456",
            "2345678234567",
        ],
        {},
    )
    assert mock_fhem.readings["testdevice"]["state"] == "offline"

    new_set = await fhempy_device.Set(testhash, ["testdevice", "?"], {})
    assert new_set == (
        "Unknown argument ?, choose one of on:noArg off:noArg "
        "countdown_1:slider,0,1,86400 relay_status:off,on,memory "
        "cycle_time random_time switch_overcharge:on,off"
    )

    # wait for readings
    await asyncio.sleep(0.1)

    assert mock_fhem.readings["testdevice"]["cur_voltage"] == 237.1

    await fhempy_device.Set(testhash, ["testdevice", "on"], {})
    # sleep to make sure that task started
    await asyncio.sleep(0.01)
    assert testval is True
    assert testindex == 1

    await fhempy_device.Set(testhash, ["testdevice", "off"], {})
    # sleep to make sure that task started
    await asyncio.sleep(0.01)
    assert testval is False
    assert testindex == 1

    await fhempy_device.Set(testhash, ["testdevice", "countdown_1", "1000"], {})
    # sleep to make sure that task started
    await asyncio.sleep(0.01)
    assert testval == 1000
    assert testindex == 9

    await fhempy_device.Set(testhash, ["testdevice", "relay_status", "memory"], {})
    # sleep to make sure that task started
    await asyncio.sleep(0.01)
    assert testval == "memory"
    assert testindex == 38

    await fhempy_device.Undefine(testhash)



from .base import BaseDevice


class HTSensor(BaseDevice):
    def __init__(self, logger, gateway):
        super().__init__(logger, gateway)


class MotionSensor(BaseDevice):
    def __init__(self, logger, gateway):
        super().__init__(logger, gateway)


class ContactSensor(BaseDevice):
    def __init__(self, logger, gateway):
        super().__init__(logger, gateway)



import asyncio
import json
import ast
import re
import tinytuya
import functools

from .. import fhem, utils
from ..generic import FhemModule
from . import mappings


class tuya(FhemModule):
    def __init__(self, logger):
        super().__init__(logger)
        self.tt_device = None
        self.tt_key = None
        self.tt_secret = None
        self.last_status = None
        self.update_lock = asyncio.Lock()

    # FHEM FUNCTION
    async def Define(self, hash, args, argsh):
        await super().Define(hash, args, argsh)
        if len(args) < 7:
            return (
                "Usage: define wifi_plug PythonModule tuya"
                " setup <API_KEY> <API_SECRET> <DEVICE_ID> [<REGION>=eu]<br>"
                "OR if you want to define only one device with existing local key<br>"
                " <PRODUCTID> <DEVICE_ID> <IP> <LOCAL_KEY> "
                "[<VERSION>=3.3] [<API_KEY>] [<API_SECRET>]"
            )

        self.tt_type = args[3]
        if self.tt_type == "setup":
            self.tt_key = args[4]
            self.tt_secret = args[5]
            self.tt_did = args[6]
            if len(args) == 8:
                self.tt_region = args[7]
            else:
                self.tt_region = "eu"
            self.set_set_config({"scan_devices": {}})
            await fhem.readingsSingleUpdate(self.hash, "state", "ready", 1)
            # set internal
            hash["DEVICEID"] = 0
            hash["API_KEY"] = self.tt_key
            hash["API_SECRET"] = self.tt_secret
            hash["REGION"] = self.tt_region
        else:
            await fhem.readingsSingleUpdateIfChanged(self.hash, "state", "offline", 1)
            self.tt_region = "eu"
            self.tt_did = args[4]
            self.tt_ip = args[5]
            self.tt_localkey = args[6]
            if len(args) >= 8:
                self.tt_version = float(args[7])
            else:
                self.tt_version = 3.3
            if len(args) == 10:
                self.tt_key = args[8]
                self.tt_secret = args[9]
            # set internal
            hash["DEVICEID"] = self.tt_did
            # set attributes
            attr_config = {
                "interval": {
                    "default": 15,
                    "format": "int",
                    "help": "Change status update interval, default is 15.",
                },
                "tuya_spec_functions": {"default": ""},
                "tuya_spec_status": {"default": ""},
            }
            self.set_attr_config(attr_config)
            # this is needed to set default values
            await utils.handle_define_attr(attr_config, self, hash)
            # create device
            await self.create_device()
            self.create_async_task(self.update_loop())

    async def _create_mapping_dev(self):
        self.tuya_spec_functions = []
        self.tuya_spec_status = []
        schema = mappings.knownSchemas[self.tt_type]["schema"]
        for schema_part in schema:
            # rename to values
            schema_part["values"] = schema_part["property"]
            schema_part["type"] = schema_part["values"]["type"]
            schema_part["desc"] = ""
            schema_part["values"] = json.dumps(schema_part["values"])
            if schema_part["type"] == "bool":
                schema_part["type"] = "Boolean"
            elif schema_part["type"] == "value":
                schema_part["type"] = "Integer"
            elif schema_part["type"] == "enum":
                schema_part["type"] = "Enum"
            elif schema_part["type"] == "string":
                schema_part["type"] = "String"
            else:
                continue
            self.tuya_spec_status.append(schema_part)
            if schema_part["mode"] == "rw":
                self.tuya_spec_functions.append(schema_part)

        await self._generate_set()

    async def set_attr_dp(self, hash):
        # check defined dp_Xs and their value
        # set set_conf based on the set dp_Xs
        dp_conf = {}
        for attr_name in self.__dict__:
            if attr_name.find("_attr_dp_") != -1:
                dp_id = re.search(r"\d+$", attr_name)[0]
                dp_conf[dp_id] = self.__dict__[attr_name]
                for fct in self.tuya_spec_functions:
                    if fct["code"] == dp_conf[dp_id]:
                        fct["id"] = int(dp_id)
                for status in self.tuya_spec_status:
                    if status["code"] == dp_conf[dp_id]:
                        status["id"] = int(dp_id)

        await self._generate_set()

    async def _generate_set(self):
        set_conf = {}
        for fct in self.tuya_spec_functions:
            if "id" not in fct:
                # no id configured for this function
                continue
            if fct["type"] == "Boolean":
                if fct["code"] == "switch_1":
                    set_conf["on"] = {
                        "help": fct["desc"],
                        "function_param": fct,
                        "function": "set_boolean",
                    }
                    set_conf["off"] = {
                        "help": fct["desc"],
                        "function_param": fct,
                        "function": "set_boolean",
                    }
                else:
                    set_conf[fct["code"]] = {
                        "options": "on,off",
                        "args": ["onoff"],
                        "help": fct["desc"],
                        "function_param": fct,
                        "function": "set_boolean",
                    }
            elif fct["type"] == "Enum":
                options = json.loads(fct["values"])["range"]
                set_conf[fct["code"]] = {
                    "options": ",".join(options),
                    "args": ["selected_val"],
                    "help": fct["desc"],
                    "function_param": fct,
                    "function": "set_enum",
                }
            elif fct["type"] == "Integer":
                spec = json.loads(fct["values"])
                slider = f"slider,{spec['min']},{spec['step']},{spec['max']}"
                set_conf[fct["code"]] = {
                    "options": slider,
                    "args": ["selected_val"],
                    "params": {"selected_val": {"format": "int"}},
                    "help": fct["desc"],
                    "function_param": fct,
                    "function": "set_integer",
                }
            elif fct["type"] == "String":
                set_conf[fct["code"]] = {
                    "args": ["new_val"],
                    "help": fct["desc"],
                    "function_param": fct,
                    "function": "set_string",
                }
            elif fct["type"] == "Json":
                pass
        self.set_set_config(set_conf)

    async def set_boolean(self, hash, params):
        switch_id = params["function_param"]["id"]
        onoff = False
        if "onoff" in params:
            if params["onoff"] == "on":
                onoff = True
        else:
            if params["cmd"] == "on":
                onoff = True
        func = self.tt_device.set_status
        self.create_async_task(
            self._call_fct_upd_reading(functools.partial(func, onoff, switch=switch_id))
        )

    async def set_integer(self, hash, params):
        index = params["function_param"]["id"]
        new_val = params["selected_val"]
        func = self.tt_device.set_value
        self.create_async_task(
            self._call_fct_upd_reading(functools.partial(func, index, new_val))
        )

    async def set_enum(self, hash, params):
        index = params["function_param"]["id"]
        new_val = params["selected_val"]
        func = self.tt_device.set_status
        self.create_async_task(
            self._call_fct_upd_reading(functools.partial(func, new_val, index))
        )

    async def set_string(self, hash, params):
        index = params["function_param"]["id"]
        new_val = params["new_val"]
        func = self.tt_device.set_status
        self.create_async_task(
            self._call_fct_upd_reading(functools.partial(func, new_val, index))
        )

    # check if tuya_spec_functions and tuya_spec_status attr is set?
    async def check_tuya_attributes(self):
        self.tuya_spec_functions = await fhem.AttrVal(
            self.hash["NAME"], "tuya_spec_functions", "[]"
        )
        self.tuya_spec_status = await fhem.AttrVal(
            self.hash["NAME"], "tuya_spec_status", "[]"
        )
        self.tuya_spec_functions = ast.literal_eval(self.tuya_spec_functions)
        self.tuya_spec_status = ast.literal_eval(self.tuya_spec_status)

    # get functions/status from tuya
    async def get_tuya_dev_specification(self, token):
        # Get specification
        uri = f"devices/{self.tt_did}/specifications"
        response_dict = await self._request_tuyacloud(uri, token)
        return response_dict["result"]

    async def get_tuya_dev_description(self, token):
        # Get function description
        uri = f"devices/{self.tt_did}/functions"
        response_dict = await self._request_tuyacloud(uri, token)
        fct_desc = response_dict["result"]["functions"]
        fct_code_desc = {}
        for fct in fct_desc:
            fct_code_desc[fct["code"]] = fct["desc"]
        return fct_code_desc

    async def _add_desc_to_spec(self, spec_fcts, desc):
        for spec in spec_fcts:
            if spec["code"] in desc:
                spec["desc"] = desc[spec["code"]]
        return spec_fcts

    async def _create_cloudmapping_dev(self):
        if self.tt_key == "" or self.tt_secret == "":
            await fhem.readingsSingleUpdateIfChanged(
                self.hash,
                "state",
                "Please use API_KEY and API_SECRET",
            )
            return

        await fhem.readingsSingleUpdate(self.hash, "state", "Initializing...", 1)
        # check if attributes are set
        await self.check_tuya_attributes()
        if len(self.tuya_spec_functions) == 0 and len(self.tuya_spec_status) == 0:
            # retrieve cloud codes
            token = await self._obtain_token()
            if token is None:
                await fhem.readingsSingleUpdateIfChanged(
                    self.hash, "state", "Login failed", 1
                )
                return

            spec = await self.get_tuya_dev_specification(token)
            self.tuya_spec_functions = spec["functions"]
            self.tuya_spec_status = spec["status"]
            desc = await self.get_tuya_dev_description(token)
            self.tuya_spec_functions = await self._add_desc_to_spec(
                self.tuya_spec_functions, desc
            )
            await fhem.CommandAttr(
                self.hash,
                (
                    f"{self.hash['NAME']} tuya_spec_functions "
                    f"{str(self.tuya_spec_functions)}"
                ),
            )
            await fhem.CommandAttr(
                self.hash,
                f"{self.hash['NAME']} tuya_spec_status {str(self.tuya_spec_status)}",
            )

        await fhem.readingsSingleUpdate(
            self.hash, "state", "Localscan...please wait...", 1
        )
        # retrieve local dp_s
        self.local_dev = None
        devices = await utils.run_blocking(
            functools.partial(tinytuya.deviceScan, False, 20)
        )

        for ip in devices:
            if devices[ip]["gwId"] == self.tt_did:
                self.local_dev = devices[ip]

        # create attributes dp_1...X
        # add options to attributes to select cloud codes
        # e.g.
        # dp_1 = "switch_1"
        # dp_42 = "switch_overcharge"
        if self.local_dev is not None:
            dps = []
            for dp in self.local_dev["dps"]["dps"]:
                dps.append(f"dp_{int(dp):02d}")

            options = []
            for opt in self.tuya_spec_status:
                options.append(opt["code"])
            attr_conf = {}
            for dp in dps:
                attr_conf[dp] = {
                    "function": "set_attr_dp",
                    "default": "",
                    "options": ",".join(options),
                }
            self._conf_attr.update(attr_conf)
            self.set_attr_config(self._conf_attr)
            await utils.handle_define_attr(attr_conf, self, self.hash)

        await fhem.readingsSingleUpdate(
            self.hash, "state", "Ready to configure dp_ attributes", 1
        )

    async def create_device(self):
        self.tt_device = tinytuya.Device(self.tt_did, self.tt_ip, self.tt_localkey)
        self.tt_device.set_version(self.tt_version)
        # doesn't seem to work as expected
        # self.tt_device.set_socketPersistent(True)
        if self.tt_type in mappings.knownSchemas:
            await self._create_mapping_dev()
        else:
            self.create_async_task(self._create_cloudmapping_dev())

    async def update_loop(self):
        while True:
            await self.update_once()
            await asyncio.sleep(self._attr_interval)

    async def update_once(self):
        async with self.update_lock:
            try:
                status = await utils.run_blocking(
                    functools.partial(self.tt_device.status)
                )
                if self.last_status != status:
                    await self.update_readings(status)
            except (ValueError, TypeError):
                pass
            except Exception:
                await fhem.readingsSingleUpdate(self.hash, "state", "offline", 1)
                self.logger.error("Failed to get current status from device")

    def convert(self, value, schema):
        if schema["type"] == "Integer":
            values = json.loads(schema["values"])
            return value / (10 ** values["scale"])
        elif schema["type"] == "Boolean":
            if value is True:
                return "on"
            return "off"
        return value

    async def update_readings(self, status):
        status = status["dps"]
        await fhem.readingsBeginUpdate(self.hash)
        try:
            stateused = False
            for dp in status:
                found = False
                for st in self.tuya_spec_status:
                    if "id" in st and st["id"] == int(dp):
                        found = True
                        reading = st["code"]
                        if st["code"] == "switch_1":
                            reading = "state"
                            stateused = True
                        await fhem.readingsBulkUpdateIfChanged(
                            self.hash,
                            reading,
                            self.convert(status[dp], st),
                        )
                        break

                if not found:
                    await fhem.readingsBulkUpdateIfChanged(
                        self.hash,
                        f"dp_{int(dp):02d}",
                        status[dp],
                    )
            if not stateused:
                await fhem.readingsBulkUpdateIfChanged(self.hash, "state", "online")
        except:
            self.logger.exception("Failed to update readings")
        await fhem.readingsEndUpdate(self.hash, 1)

    async def _call_fct_upd_reading(self, function):
        try:
            await utils.run_blocking(function)
        except:
            self.logger.exception("Failed to execute function")
        await self.update_once()

    async def Undefine(self, hash):
        if self.tt_device:
            del self.tt_device
        await super().Undefine(hash)

    # The following code is only for setup/scan process via tuya cloud
    async def set_scan_devices(self, hash, params):
        self.create_async_task(self._scan_devices())

    async def _request_tuyacloud(self, uri, token=None):
        return await utils.run_blocking(
            functools.partial(
                tinytuya.tuyaPlatform,
                self.tt_region,
                self.tt_key,
                self.tt_secret,
                uri,
                token,
            )
        )

    async def _obtain_token(self):
        # Get Oauth Token from tuyaPlatform
        uri = "token?grant_type=1"
        response_dict = await self._request_tuyacloud(uri)
        if "result" in response_dict:
            token = response_dict["result"]["access_token"]
            return token
        return None

    async def _scan_devices(self):
        token = await self._obtain_token()

        # Get UID from sample Device ID
        uri = "devices/%s" % self.tt_did
        response_dict = await self._request_tuyacloud(uri, token)
        uid = response_dict["result"]["uid"]

        # Use UID to get list of all Devices for User
        uri = "users/%s/devices" % uid
        json_data = await self._request_tuyacloud(uri, token)

        # Filter to only Name, ID and Key
        tuyadevices = []
        for i in json_data["result"]:
            item = {}
            item["name"] = i["name"].strip()
            item["id"] = i["id"]
            item["key"] = i["local_key"]
            item["product_id"] = i["product_id"]
            item["icon"] = i["icon"]
            await fhem.readingsBeginUpdate(self.hash)
            await fhem.readingsBulkUpdateIfChanged(
                self.hash, item["id"] + "_name", item["name"]
            )
            await fhem.readingsBulkUpdateIfChanged(
                self.hash, item["id"] + "_id", item["id"]
            )
            await fhem.readingsBulkUpdateIfChanged(
                self.hash, item["id"] + "_localkey", item["key"]
            )
            await fhem.readingsBulkUpdateIfChanged(
                self.hash, item["id"] + "_productid", item["product_id"]
            )
            await fhem.readingsBulkUpdateIfChanged(
                self.hash,
                item["id"] + "_icon",
                "https://images.tuyaeu.com/" + item["icon"],
            )
            await fhem.readingsEndUpdate(self.hash, 1)
            tuyadevices.append(item)

        # Display device list
        self.logger.debug("Device Listing")
        output = json.dumps(tuyadevices, indent=4)  # sort_keys=True)
        self.logger.debug(output)
        await fhem.readingsSingleUpdate(
            self.hash,
            "state",
            f"found {len(tuyadevices)} devices, start local scan...",
            1,
        )

        # scan local devices to get IP
        self.logger.debug("Scan local devices...")
        devices = await utils.run_blocking(
            functools.partial(tinytuya.deviceScan, False, 20)
        )

        def getIP(d, gwid):
            for ip in d:
                if gwid == d[ip]["gwId"]:
                    return (ip, d[ip]["version"])
            return (0, 0)

        self.logger.debug("Polling local devices...")
        count_created = 0
        for i in tuyadevices:
            name = i["name"].replace(" ", "_")
            id = i["id"]
            (ip, ver) = getIP(devices, i["id"])
            local_key = i["key"]
            productid = i["product_id"]

            # do not create device if no IP was discovered
            if ver == 0:
                await fhem.readingsSingleUpdateIfChanged(
                    self.hash, f"{id}_ip", "offline", 1
                )
                continue

            await fhem.readingsSingleUpdateIfChanged(self.hash, f"{id}_ip", ip, 1)
            await fhem.readingsSingleUpdateIfChanged(
                self.hash, f"{id}_version", str(ver), 1
            )

            # do not create a device if unknown
            if productid not in mappings.knownSchemas:
                continue

            if not await fhem.checkIfDeviceExists(
                self.hash, "PYTHONTYPE", "tuya", "DEVICEID", id
            ):
                # create FHEM device
                await fhem.CommandDefine(
                    self.hash,
                    (
                        f"{name}_{id} PythonModule tuya "
                        f"{productid} {id} {ip} {local_key} "
                        f"{ver} {self.tt_key} {self.tt_secret}"
                    ),
                )
                count_created += 1

        await fhem.readingsSingleUpdate(
            self.hash,
            "state",
            f"done, created {count_created} devices",
            1,
        )
